/*
 * StarPU
 * Copyright (C) Université Bordeaux 1, CNRS 2008-2010 (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU Lesser General Public License in COPYING.LGPL for more details.
 */

#include <pthread.h>
#include <core/mechanisms/fifo_queues.h>
#include <errno.h>
#include <common/utils.h>
#include <core/task.h>

struct starpu_fifo_jobq_s *_starpu_create_fifo(void)
{
	struct starpu_fifo_jobq_s *fifo;
	fifo = malloc(sizeof(struct starpu_fifo_jobq_s));

	/* note that not all mechanisms (eg. the semaphore) have to be used */
	fifo->jobq = starpu_job_list_new();
	fifo->njobs = 0;
	fifo->nprocessed = 0;

	fifo->exp_start = _starpu_timing_now();
	fifo->exp_len = 0.0;
	fifo->exp_end = fifo->exp_start;

	return fifo;
}

void _starpu_destroy_fifo(struct starpu_fifo_jobq_s *fifo)
{
	starpu_job_list_delete(fifo->jobq);
	free(fifo);
}

int _starpu_fifo_push_prio_task(struct starpu_fifo_jobq_s *fifo_queue, pthread_mutex_t *sched_mutex, pthread_cond_t *sched_cond, struct starpu_task *task)
{
	starpu_job_t j = _starpu_get_job_associated_to_task(task);

	PTHREAD_MUTEX_LOCK(sched_mutex);

	STARPU_TRACE_JOB_PUSH(task, 0);
	starpu_job_list_push_back(fifo_queue->jobq, j);
	fifo_queue->njobs++;
	fifo_queue->nprocessed++;

	pthread_cond_signal(sched_cond);
	PTHREAD_MUTEX_UNLOCK(sched_mutex);

	return 0;
}

int _starpu_fifo_push_task(struct starpu_fifo_jobq_s *fifo_queue, pthread_mutex_t *sched_mutex, pthread_cond_t *sched_cond, struct starpu_task *task)
{
	starpu_job_t j = _starpu_get_job_associated_to_task(task);

	PTHREAD_MUTEX_LOCK(sched_mutex);

	STARPU_TRACE_JOB_PUSH(task, 0);
	starpu_job_list_push_front(fifo_queue->jobq, j);
	fifo_queue->njobs++;
	fifo_queue->nprocessed++;

	pthread_cond_signal(sched_cond);
	PTHREAD_MUTEX_UNLOCK(sched_mutex);

	return 0;
}

struct starpu_task *_starpu_fifo_pop_task(struct starpu_fifo_jobq_s *fifo_queue)
{
	starpu_job_t j = NULL;

	if (fifo_queue->njobs == 0)
		return NULL;

	if (fifo_queue->njobs > 0) 
	{
		/* there is a task */
		j = starpu_job_list_pop_back(fifo_queue->jobq);
	
		STARPU_ASSERT(j);
		fifo_queue->njobs--;
		
		STARPU_TRACE_JOB_POP(j, 0);
	}
	
	return j->task;
}

/* pop every task that can be executed on the calling driver */
struct starpu_task_list *_starpu_fifo_pop_every_task(struct starpu_fifo_jobq_s *fifo_queue, pthread_mutex_t *sched_mutex, uint32_t where)
{
	struct starpu_job_list_s *old_list;
	unsigned size;

	struct starpu_task_list *new_list = NULL;
	struct starpu_task_list *new_list_tail = NULL;
	
	PTHREAD_MUTEX_LOCK(sched_mutex);

	size = fifo_queue->njobs;

	if (size > 0) {
		old_list = fifo_queue->jobq;
		unsigned new_list_size = 0;

		starpu_job_itor_t i;
		starpu_job_t next_job;
		/* note that this starts at the _head_ of the list, so we put
 		 * elements at the back of the new list */
		for(i = starpu_job_list_begin(old_list);
			i != starpu_job_list_end(old_list);
			i  = next_job)
		{
			next_job = starpu_job_list_next(i);

			if (i->task->cl->where & where)
			{
				/* this elements can be moved into the new list */
				new_list_size++;
				
				starpu_job_list_erase(old_list, i);

				if (new_list)
				{
					struct starpu_task_list *link;

					link = malloc(sizeof(struct starpu_task_list));
					link->task = i->task;
					link->next = NULL;

					new_list_tail->next = link;
					new_list_tail = link;
					
				}
				else {
					new_list = malloc(sizeof(struct starpu_task_list));
					new_list->task = i->task;
					new_list->next = NULL;
					new_list_tail = new_list;
				}
			}
		}

		fifo_queue->njobs -= new_list_size;
	}

	PTHREAD_MUTEX_UNLOCK(sched_mutex);

	return new_list;
}
